#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "m_data.h"
#include "m_sr.h"

/* Save, restore routines */

void SaveMusStatus(struct MusData *Buf)
{
    int p = MusData.Paused;
    SetPause(1);
    FreeUncompressedData();
    memcpy(Buf, &MusData, sizeof(*Buf));
    /* FIXME: Save music device status */
    SetPause(p);
}

void RestoreMusStatus(const struct MusData *Buf)
{
    SetPause(1);
    memcpy(&MusData, Buf, sizeof(*Buf));
    if(MusData.driver)MusData.driver->reset();
    UncompressChannels();
    /* Paused flag was saved as 1 */
    SetPause(0);
}

void FreeUncompressedData(void)
{
    unsigned a;
    for(a=0; a<MAXCHN; ++a)
    {
#ifdef COMPRESSED_CHANNELS
        if(MusData.PlayData.ch[a].patndata)
            free(MusData.PlayData.ch[a].patndata);
#endif
    }
}

#include "bitfun.h"

void UncompressChannels(void)
{
    unsigned a;
    
    const struct InternalChanData *chans = MusData.SongData.Chans[MusData.PlayData.posi.Pat];
    
    printf("Pattern %u:\n", MusData.PlayData.posi.Pat);
    
    for(a=0; a<MusData.SongData.Hdr->ChanCount; ++a)
    {
        const BYTE *src = (const BYTE *)chans + sizeof(*chans);
        unsigned chansize = (chans->BitCount+7) >> 3; /* Channel size in bytes */
#ifdef COMPRESSED_CHANNELS
        BYTE *dest = (BYTE *)calloc(chans->OriginalByteSize, 1); /* We want zero */
        unsigned srcbitpos=0, destbitpos=0;
        
        const BYTE bbb = ReadBits(src, srcbitpos,   4);
        const BYTE ooo = ReadBits(src, srcbitpos+4, 4);
        const BYTE ppp=5+ReadBits(src, srcbitpos+8, 1);
        const BYTE limit = 1 + bbb + ooo;
        
        srcbitpos += 4+4+1;
        
        MusData.PlayData.ch[a].patndata = dest;
        
        printf("  Channel=%u: bbb=%u, ooo=%u, ppp=%u - size=%u (%u bits)\n",
            a, bbb, ooo, ppp, chansize, chans->BitCount);
        
        /* DWORD ReadBits(const BYTE *data, unsigned bitpos, unsigned bitcount)       */
        /* void WriteBits(BYTE *dest, unsigned bitpos, unsigned bitcount, DWORD data) */

        for(; srcbitpos < chans->BitCount; )
        {
            BYTE k = ReadBits(src, srcbitpos++, 1);
            if(k)
            {
                WORD bb = ReadBits(src, srcbitpos,       bbb);
                WORD oo = ReadBits(src, srcbitpos + bbb, ooo);
                unsigned targetpos = destbitpos-oo-1;
                unsigned size      = bb+limit+1;
                srcbitpos += bbb + ooo;
                
                printf("   Copy %u(%u) from %u(%u) at %u\n",
                    bb,size,
                    oo,targetpos, destbitpos);
                
                CopyBits(dest, destbitpos, dest, targetpos, size);
                destbitpos += size;
            }
            else
            {
                WORD pp = ReadBits(src, srcbitpos, ppp);
                srcbitpos += ppp;
                unsigned size=pp;
                
                printf("   Raw*%u(%u) at %u\n", pp,size, destbitpos);
                
                CopyBits(dest, destbitpos, src, srcbitpos, size);
                destbitpos += size;
                srcbitpos  += size;
            }
        }
        
        printf("   - Result %u bits (%u bytes - should be %u)\n",
            destbitpos,
            (destbitpos+7) >> 3,
            chans->OriginalByteSize);
#if 0
        /*if(Verbose>4)*/
        {
            unsigned len = chans->OriginalByteSize;
            unsigned b,c=0;
            printf("Chan %u:\n", a);
            for(b=0; b<len; ++b)
            {
                if(c == 30){c=0;printf("\n");}++c;
                printf(" %02X", dest[b]);
            }
            if(c)printf("\n");
        }
#endif
#else
        /* uncompressed channels */
        MusData.PlayData.ch[a].patndata = src;
#endif
        chans = (const struct InternalChanData *)(src + chansize);
    }
}
